#include <alloca.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>
#include "tools.h"

#define EXZESS_VENDOR ((uint16_t)0x0815)
#define EXZESS_DEVICE ((uint16_t)0x4711)

#define XILINX_VENDOR ((uint16_t)0x10ee)
#define XILINX_DEVICE ((uint16_t)0x0007)

#define ROUNDS 100

static void memset_addr(void *buf, int ignored, size_t size)
{
	int i;
	int *data = buf;
	size -= size % sizeof(void *);

	for (int i = 0; i < size / sizeof(void *); i++)
		data[i] = i;
}

/* The same as above, just with (pseudo-)randomized addresses. */
static void random_memset_addr(void *buf, int ignored, size_t size)
{
	int i;
	int *data = buf;
	size -= size % sizeof(void *);

	for (int i = 0; i < size / sizeof(void *); i++)
		data[random() % (size / sizeof(void *))] = i;
}


/* Just a stupid mem_read that does nothing sensible but forces the compiler to read everything.
 * Useless perhaps because volatile should do that anyways. */
static int32_t sum_int32(int32_t *buf, size_t size)
{
	int i; 
	int32_t res = 0, *data = buf;
	size -= size % sizeof(int32_t);

	for (int i = 0; i < size / sizeof(int32_t); i++)
		res += data[i];

	return res;
}

static int32_t random_sum_int32(int32_t *buf, size_t size)
{
	int i; 
	int32_t res = 0, *data = buf;
	size -= size % sizeof(int32_t);

	for (int i = 0; i < size / sizeof(int32_t); i++)
		res += data[random() % (size / sizeof(void *))];

	return res;
}

int main(int argc, char **argv)
{
	struct timespec start, stop;

	size_t bar1_size;

	set_log_level(LL_TRACE);
	char *dev_dir = find_pci_device_dir(EXZESS_VENDOR, EXZESS_DEVICE);
	if (! dev_dir) {
		dev_dir = find_pci_device_dir(XILINX_VENDOR, XILINX_DEVICE);
	}
	exit_if (NULL == dev_dir, "no suitable device found\n");
	log_msg(LL_NOTICE, "found device at %s\n", dev_dir);

	perror_if(-1 == chdir(dev_dir), "could not cd to %s", dev_dir);
	int resfd0 = open("resource0", O_RDWR);
	perror_if(-1 == resfd0, "could not open resource0 file: ");

	uint32_t * volatile bar0 = mmap(NULL, 1024, PROT_READ | PROT_WRITE, MAP_SHARED, resfd0, 0);
	perror_if(MAP_FAILED == bar0, "mmap of BAR 0 failed\n");

	int resfd1 = open("resource1", O_RDWR);
	perror_if(-1 == resfd1, "could not open resource1 file: ");

	print_exzess_status(bar0);
	printf("hexdump of BAR0 area:\n");
	printf("%s", hexdump(bar0, 0x80));

	// hopefully, the kernel did the right thing
	bar1_size = be32toh(bar0[2]);

	// but sanity-checks are good...
	if (bar1_size < 2 || bar1_size > 1000000) {
		printf("got bar1_size of %d, and I do not like it, exiting.\n");
		exit(EXIT_FAILURE);
	}

	uint32_t * volatile bar1 = mmap(NULL, bar1_size, PROT_READ | PROT_WRITE, MAP_SHARED, resfd1, 0);
	perror_if(MAP_FAILED == bar1, "mmap of BAR 1 failed\n");
	
	printf("hexdump of BAR0 area:\n");
	printf("%s", hexdump(bar1, 0x80));

	perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start), "clock_gettime start");
	memset(bar1, 0x00, bar1_size);
	perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &stop), "clock_gettime stop");

	printf("hexdump of BAR1 area:\n");
	printf("%s", hexdump(bar1, 0x80));

	printf("memset 0 speed: %.2lf kB/s\n", 0.001 * calculate_bytes_per_sec(&start, &stop, bar1_size));

	double wr_res[ROUNDS];
	double rd_res[ROUNDS];
	int32_t sum = 0;
	for (int i = 0; i < ROUNDS; i++) {
		perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start), "clock_gettime start");
		memset_addr(bar1, 0xaa, bar1_size);
		perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &stop), "clock_gettime stop");

		wr_res[i] = calculate_bytes_per_sec(&start, &stop, bar1_size);

		perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start), "clock_gettime start");
		sum += sum_int32(bar1, bar1_size);
		perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &stop), "clock_gettime stop");

		rd_res[i] = calculate_bytes_per_sec(&start, &stop, bar1_size);
	}

	printf("%d rounds of memset 0 avg speed: %.2lf kB/s stddev: %.2lf kB/s\n", ROUNDS, 0.001 * avg(wr_res, ROUNDS), 0.001 * stddev(wr_res, ROUNDS));
	printf("%d rounds of sum (%d) avg speed: %.2lf kB/s stddev: %.2lf kB/s\n", ROUNDS, sum, 0.001 * avg(rd_res, ROUNDS), 0.001 * stddev(rd_res, ROUNDS));

	for (int i = 0; i < ROUNDS; i++) {
		srandom(i + 3141592);

		perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start), "clock_gettime start");
		random_memset_addr(bar1, 0xaa, bar1_size);
		perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &stop), "clock_gettime stop");

		wr_res[i] = calculate_bytes_per_sec(&start, &stop, bar1_size);

		perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &start), "clock_gettime start");
		sum += random_sum_int32(bar1, bar1_size);
		perror_if(-1 == clock_gettime(CLOCK_PROCESS_CPUTIME_ID, &stop), "clock_gettime stop");

		rd_res[i] = calculate_bytes_per_sec(&start, &stop, bar1_size);
	}

	printf("%d rounds of randomized memset 0 avg speed: %.2lf kB/s stddev: %.2lf kB/s\n", ROUNDS, 0.001 * avg(wr_res, ROUNDS), 0.001 * stddev(wr_res, ROUNDS));
	printf("%d rounds of randomized sum (%d) avg speed: %.2lf kB/s stddev: %.2lf kB/s\n", ROUNDS, sum, 0.001 * avg(rd_res, ROUNDS), 0.001 * stddev(rd_res, ROUNDS));

	memset(bar1, 0x00, bar1_size);
	exit(EXIT_SUCCESS);
}
