/*
 *  st -- struct tool
 * 
 *  Copyright 2010-2011 Alexander Würstlein <arw@cs.fau.de>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <assert.h>
#include <ctype.h>
#include <dirent.h>
#include <endian.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>
#include <unistd.h>

#include "tools.h"

static enum log_level log_level = LL_WARNING;
static FILE *log_fd = NULL;

static void log_va(enum log_level level, char *position, char *fmt, va_list ap)
{
	char *msg;
	if (NULL == log_fd) {
		log_fd = stderr;
	};
	if (level > log_level) {
		if (log_level > LL_WARNING);
			fprintf(log_fd, position);

		vasprintf(&msg, fmt, ap);
		fprintf(log_fd, msg);
		free(msg);
	};
}

void _log_msg(enum log_level level, char *position, char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	log_va(level, position, fmt, ap);
	va_end(ap);
}

void _log_if(int cond, enum log_level level, char *position, char *fmt, ...)
{
	va_list ap;
	if (cond) {
		va_start(ap, fmt);
		log_msg(level, position, fmt, ap);
		va_end(ap);
	}
}

void set_log_level(enum log_level level)
{
	log_level = level;
}

enum log_level get_log_level()
{
	return log_level;
}

void _exit_if(int cond, char *position, char *fmt, ...)
{
	va_list ap;
	if (cond) {
		va_start(ap, fmt);
		log_va(LL_ERROR, position, fmt, ap);
		va_end(ap);
		exit(EXIT_FAILURE);
	}
}

void _perror_if(int cond, char *position, char *fmt, ...)
{
	va_list ap;
	if (cond) {
		va_start(ap, fmt);
		log_va(LL_ERROR, position, fmt, ap);
		va_end(ap);
		_log_msg(LL_ERROR, "", strerror(errno));
		exit(EXIT_FAILURE);
	}
}

static char makeprintable(char c)
{
	return isprint(c)?c:'.';
}

char *hexdump(void *stuff, size_t size)
{
	int line, byte;
	char *data;
	
	data = malloc(size);
	perror_if(NULL == data, "malloc failed");
	memcpy(data, stuff, size);
	/* pro byte: 2 byte hex, 1 space, 1 ascii */
	/* pro zeile: 16 byte addr, 6 spaces, 2 |, newline, reserve/nullbyte */
	char *res = malloc(size * 4 + (1 + size / 16) * (16 + 6 + 2 + 3));
	perror_if(NULL == res, "malloc failed");
	char *cur;
	cur = res;
	int n;

	for (line = 0; line < size / 16; line++) {
		n = sprintf(cur, "%016x  %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx  %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx %02hhx  |%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c%c|\n", 
				((int)data + (line * 16)), 
				data[0], data[1], data[2], data[3],
				data[4], data[5], data[6], data[7],
				data[8], data[9], data[10], data[11],
				data[12], data[13], data[14], data[15],
				makeprintable(data[0]), makeprintable(data[1]), makeprintable(data[2]), makeprintable(data[3]),
				makeprintable(data[4]), makeprintable(data[5]), makeprintable(data[6]), makeprintable(data[7]),
				makeprintable(data[8]), makeprintable(data[9]), makeprintable(data[10]), makeprintable(data[11]),
				makeprintable(data[12]), makeprintable(data[13]), makeprintable(data[14]), makeprintable(data[15])
			);
		perror_if(n <  16 + 16 * 3 + 6 + 2 + 1, "wrong number of output characters: %d", n);
		cur += n;
		data += 16;
	}
	if (size % 16 != 0) {
		n = sprintf(cur, "%016x  ", ((int) data + (line * 16)));
		perror_if(n != 18, "wrong number of output characters");
		cur += n;
		for (byte = 0; byte < size % 16; byte++)
		{
			n = sprintf(cur, "%02x ", (unsigned int)data[byte]);
			perror_if(n != 3, "wrong number of output characters");
			cur += n;
			if (byte == 7) {
				n = sprintf(cur, " ");
				perror_if(n != 1, "wrong number of output characters");
				cur += n;
			}
		}
		for (byte = 0; byte < (48 - 3 * (size % 16)); byte++) {
			n = sprintf(cur, " ");
			perror_if(n != 1, "wrong number of output characters");
			cur += n;
		}

		if (size % 16 < 8) {
			n = sprintf(cur, " ");
			perror_if(n != 1, "wrong number of output characters");
			cur += n;
		}

		n = sprintf(cur, " |");
		perror_if(n != 2, "wrong number of output characters");
		cur += n;

		for (byte = 0; byte < size % 16; byte++) {
			n = sprintf(cur, "%c", makeprintable(data[byte]));
			perror_if(n != 1, "wrong number of output characters");
			cur += n;
		}

		for (byte = 0; byte < 16 - (size % 16); byte++) {
			n = sprintf(cur, " ");
			perror_if(n != 1, "wrong number of output characters");
			cur += n;
		}

		n = sprintf(cur, "|\n");
		perror_if(n != 2, "wrong number of output characters");
		cur += n;
	}
	return res;
}

char *find_pci_device_dir(uint16_t vendor, uint16_t device)
{
	char *sysfsdir = "/sys/bus/pci/devices";
	char *res = NULL;
	DIR *d;
	d = opendir(sysfsdir);
	char vendor_str[8];
	char device_str[8];
	sprintf(vendor_str, "0x%04x", vendor);
	sprintf(device_str, "0x%04x", device);

	exit_if(NULL == d, "could not open %s", sysfsdir);

	struct dirent *de;

	while (de = readdir(d)) {
		if (strcmp(".", de->d_name) == 0 || strcmp("..", de->d_name) == 0)
			continue;

		res = malloc(strlen(sysfsdir) + strlen(de->d_name) + 2);
		sprintf(res, "%s/%s", sysfsdir, de->d_name);

		char *vfn = alloca(strlen(sysfsdir) + strlen(de->d_name) + sizeof("vendor") + 3);
		char *dfn = alloca(strlen(sysfsdir) + strlen(de->d_name) + sizeof("device") + 3);
		sprintf(vfn, "%s/%s/%s", sysfsdir, de->d_name, "vendor");
		sprintf(dfn, "%s/%s/%s", sysfsdir, de->d_name, "device");

		errno = 0;
		int vfd = open(vfn, O_RDONLY);
		perror_if(0 == vfd, "failed to open vendor id file at %s", vfn);
		errno = 0;
		char vfs[10];
		perror_if(1 > read(vfd, vfs, 9), "read failed on vendor id file %s", vfn);
		vfs[6] = 0;

		errno = 0;
		int dfd = open(dfn, O_RDONLY);
		perror_if(0 == dfd, "failed to open device id file at %s", dfn);
		errno = 0;
		char dfs[10];
		perror_if(1 > read(dfd, dfs, 9), "read failed on vendor id file %s", dfn);
		dfs[6] = 0;

		perror_if(-1 == close(vfd) || -1 == close(dfd), "close failed");

		log_msg(LL_TRACE, "device at %s: %s:%s\n", res, vfs, dfs);

		if (strcmp(vfs, vendor_str) == 0 && strcmp(dfs, device_str) == 0) {
			/* we found it, clean up and return res */
			break;
		} else {
			/* nothing there, next one, or at the end, return res = NULL */
			free(res);
			res = NULL;
		}
	}
	closedir(d);

	return res;
}

/** 
 * read device status from bar0
 *
 * @param bar0 address where bar0 is mapped
 */

void print_exzess_status(uint32_t * const volatile bar0)
{
	exit_if(NULL == bar0, "found NULL pointer where bar0 address expected");
	printf("DMA adr: %08x, size: %d, in (rd: %d, wr: %d, cpld: %d), out (rd: %d, wr: %d, cpld: %d)\n",
			be32toh(bar0[1]), be32toh(bar0[2]), // DMA adr and size
			be32toh(bar0[8]), be32toh(bar0[9]), be32toh(bar0[10]), // incoming TLP counters
			be32toh(bar0[11]), be32toh(bar0[12]), be32toh(bar0[13]) // outgoing TLP counters
			);
}

/**
 *
 * calculate a bytes/sec value from two struct timespec and a byte count
 */
double calculate_bytes_per_sec(struct timespec * const start, struct timespec * const stop, uint64_t bytes)
{
	double sdiv, nsdiv;
	exit_if(NULL == start || NULL == stop, "can not handle NULL arguments");

	if (start->tv_sec == stop->tv_sec && start->tv_nsec == stop->tv_nsec) return INFINITY;

	// two steps to avoid losing precision, difference is more exact for int types
	sdiv = (double)(stop->tv_sec - start->tv_sec);
	nsdiv = (double)(stop->tv_nsec - start->tv_nsec);

	return ((double)bytes) / (sdiv + nsdiv / 1000000000.0);
}

double sum(double *data, int count)
{
	double res = 0.0;
	for (int i = 0; i < count; i++)
		res += data[i];
	return res;
}

double avg(double *data, int count)
{
	return sum(data, count) / (double)count;
}

double stddev(double *data, int count)
{
	double average = avg(data, count);
	double squares[count];
	for (int i = 0; i < count; i++) {
		squares[i] = data[i] - average;
		squares[i] *= squares[i];
	}

	return sqrt(sum(squares, count) / ((double)(count - 1)));
}
