/*
 *  from: st -- struct tool
 * 
 *  Copyright 2010-2011 Alexander Würstlein <arw@cs.fau.de>
 * 
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/* various useful things */
#ifndef TOOLS_H
#define TOOLS_H

#include <inttypes.h>
#include <sys/types.h>

/* 
 * typesafe min/max/clamp and container_of are from the linux kernel.h file
 */

/*
 * min()/max()/clamp() macros that also do
 * strict type-checking.. See the
 * "unnecessary" pointer comparison.
 */
#define min(x, y) ({				\
	typeof(x) _min1 = (x);			\
	typeof(y) _min2 = (y);			\
	(void) (&_min1 == &_min2);		\
	_min1 < _min2 ? _min1 : _min2; })

#define max(x, y) ({				\
	typeof(x) _max1 = (x);			\
	typeof(y) _max2 = (y);			\
	(void) (&_max1 == &_max2);		\
	_max1 > _max2 ? _max1 : _max2; })

#define min3(x, y, z) ({			\
	typeof(x) _min1 = (x);			\
	typeof(y) _min2 = (y);			\
	typeof(z) _min3 = (z);			\
	(void) (&_min1 == &_min2);		\
	(void) (&_min1 == &_min3);		\
	_min1 < _min2 ? (_min1 < _min3 ? _min1 : _min3) : \
		(_min2 < _min3 ? _min2 : _min3); })

#define max3(x, y, z) ({			\
	typeof(x) _max1 = (x);			\
	typeof(y) _max2 = (y);			\
	typeof(z) _max3 = (z);			\
	(void) (&_max1 == &_max2);		\
	(void) (&_max1 == &_max3);		\
	_max1 > _max2 ? (_max1 > _max3 ? _max1 : _max3) : \
		(_max2 > _max3 ? _max2 : _max3); })

/**
 * min_not_zero - return the minimum that is _not_ zero, unless both are zero
 * @x: value1
 * @y: value2
 */
#define min_not_zero(x, y) ({			\
	typeof(x) __x = (x);			\
	typeof(y) __y = (y);			\
	__x == 0 ? __y : ((__y == 0) ? __x : min(__x, __y)); })

/**
 * clamp - return a value clamped to a given range with strict typechecking
 * @val: current value
 * @min: minimum allowable value
 * @max: maximum allowable value
 *
 * This macro does strict typechecking of min/max to make sure they are of the
 * same type as val.  See the unnecessary pointer comparisons.
 */
#define clamp(val, min, max) ({			\
	typeof(val) __val = (val);		\
	typeof(min) __min = (min);		\
	typeof(max) __max = (max);		\
	(void) (&__val == &__min);		\
	(void) (&__val == &__max);		\
	__val = __val < __min ? __min: __val;	\
	__val > __max ? __max: __val; })

/*
 * ..and if you can't take the strict
 * types, you can specify one yourself.
 *
 * Or not use min/max/clamp at all, of course.
 */
#define min_t(type, x, y) ({			\
	type __min1 = (x);			\
	type __min2 = (y);			\
	__min1 < __min2 ? __min1: __min2; })

#define max_t(type, x, y) ({			\
	type __max1 = (x);			\
	type __max2 = (y);			\
	__max1 > __max2 ? __max1: __max2; })

/**
 * clamp_t - return a value clamped to a given range using a given type
 * @type: the type of variable to use
 * @val: current value
 * @min: minimum allowable value
 * @max: maximum allowable value
 *
 * This macro does no typechecking and uses temporary variables of type
 * 'type' to make all the comparisons.
 */
#define clamp_t(type, val, min, max) ({		\
	type __val = (val);			\
	type __min = (min);			\
	type __max = (max);			\
	__val = __val < __min ? __min: __val;	\
	__val > __max ? __max: __val; })

/**
 * clamp_val - return a value clamped to a given range using val's type
 * @val: current value
 * @min: minimum allowable value
 * @max: maximum allowable value
 *
 * This macro does no typechecking and uses temporary variables of whatever
 * type the input argument 'val' is.  This is useful when val is an unsigned
 * type and min and max are literals that will otherwise be assigned a signed
 * integer type.
 */
#define clamp_val(val, min, max) ({		\
	typeof(val) __val = (val);		\
	typeof(val) __min = (min);		\
	typeof(val) __max = (max);		\
	__val = __val < __min ? __min: __val;	\
	__val > __max ? __max: __val; })


/*
 * swap - swap value of @a and @b
 */
#define swap(a, b) \
	do { typeof(a) __tmp = (a); (a) = (b); (b) = __tmp; } while (0)

/**
 * container_of - cast a member of a structure out to the containing structure
 * @ptr:	the pointer to the member.
 * @type:	the type of the container struct this is embedded in.
 * @member:	the name of the member within the struct.
 *
 */
#define container_of(ptr, type, member) ({			\
	const typeof( ((type *)0)->member ) *__mptr = (ptr);	\
	(type *)( (char *)__mptr - offsetof(type,member) );})


enum log_level {
	LL_TRACE = 1,
	LL_DEBUG = 2,
	LL_NOTICE = 3,
	LL_INFO = 4,
	LL_WARNING = 5,
	LL_ERROR = 6,
};

/* very ugly variadic macros
 * see http://gcc.gnu.org/onlinedocs/cpp/Variadic-Macros.html for explanation of the syntax
 * perhaps this is nonportable and gcc only
 */

#define SSTR(x) #x
#define STR(x) SSTR(x)

#define log_if(cond, level, ...) _log_if(cond, level, __FILE__": "STR(__LINE__)"\t" , ##__VA_ARGS__)
#define log_msg(level, ...) _log_msg(level, __FILE__": "STR(__LINE__)"\t" , ##__VA_ARGS__)
#define exit_if(cond, ...) _exit_if(cond, __FILE__": "STR(__LINE__)"\t" , ##__VA_ARGS__)
#define perror_if(cond, ...) _perror_if(cond, __FILE__": "STR(__LINE__)"\t" , ##__VA_ARGS__)

void _log_msg(enum log_level level, char *position, char *fmt, ...);
void _log_if(int cond, enum log_level level, char *position, char *fmt, ...);
void set_log_level(enum log_level level);
enum log_level get_log_level();

void _exit_if(int cond, char *position, char *fmt, ...);
void _perror_if(int cond, char *position, char *fmt, ...);

char *get_tabs(unsigned int n);

char *hexdump(void *stuff, size_t size);
char *find_pci_device_dir(uint16_t vendor, uint16_t device);
void print_exzess_status(uint32_t * const volatile bar0);

double calculate_bytes_per_sec(struct timespec * const start, struct timespec * const stop, uint64_t bytes);
double sum(double *data, int count);
double avg(double *data, int count);
double stddev(double *data, int count);

/* generic function type, similar to 'void' for data types
 * needed because void * can't be cast into a function pointer
 */
typedef void void_fn(void);
typedef void (*void_fn_ptr)(void);

#endif /* ifndef TOOLS_H */
